from pykzee.core.Tree import Tree


class Plugin:
    __slots__ = (
        "path",
        "get",
        "subscribe",
        "command",
        "setState",
        "registerCommand",
    )

    def __init__(
        self, *, path, get, subscribe, command, set_state, register_command
    ):
        self.path = path
        self.get = get
        self.subscribe = subscribe
        self.command = command
        self.set = set_state
        self.registerCommand = register_command

    def createSubtree(self, path, *, immediate_updates=True):
        return Tree(
            path,
            parent_set=self.set,
            parent_register_command=self.registerCommand,
            immediate_updates=immediate_updates,
        )



import functools

from pyimmutable import ImmutableList, ImmutableDict
from pykzee.core.common import Undefined, makePath, pathToString

SameAsData = object()


def attribute(key):
    def decorator(func):
        def modified_func(data):
            try:
                result = data.meta[key]
            except KeyError:
                ...
            else:
                return data if result is SameAsData else result

            result = func(data)
            data.meta[key] = SameAsData if result is data else result
            return result

        modified_func.uncached = func
        return modified_func

    return decorator


##############################################################################


def getSubtree(data, path):
    if not path:
        return data
    if type(data) not in (ImmutableDict, ImmutableList):
        return Undefined

    meta = data.meta
    try:
        cache = meta["subtree-cache"]
    except KeyError:
        cache = meta["subtree-cache"] = {}

    try:
        return cache[path]
    except KeyError:
        ...

    try:
        child = data[path[0]]
    except Exception:
        cache[path] = Undefined
        return Undefined

    if len(path) < 2:
        cache[path] = child
        return child

    result = cache[path] = getSubtree(child, path[1:])
    return result


@attribute("plugins")
def plugins(data):
    if type(data) is ImmutableDict:
        if "__plugin__" in data:
            return ImmutableList([((), data)])

    gen = (
        sorted(data.items())
        if type(data) is ImmutableDict
        else enumerate(data)
    )
    return ImmutableList(
        ((key,) + path, plugin)
        for key, value in gen
        if type(value) in (ImmutableDict, ImmutableList)
        for path, plugin in plugins(value)
    )


def symlink(data):
    if type(data) is ImmutableDict and len(data) == 1:
        (key, value), = data.items()
        if key == "__symlink__":
            try:
                return makePath(value)
            except TypeError:
                return False


@attribute("symlinks")
def symlinks(data):
    sl = symlink(data)
    if sl is False:
        return ImmutableList()
    elif sl is not None:
        return ImmutableList([((), sl)])

    gen = (
        sorted(data.items())
        if type(data) is ImmutableDict
        else enumerate(data)
    )
    return ImmutableList(
        ((key,) + path, dest)
        for key, value in gen
        if type(value) in (ImmutableDict, ImmutableList)
        for path, dest in symlinks(value)
    )


@attribute("_symlinkInfoDict")
def _symlinkInfoDict(data):
    return ImmutableDict((pathToString(k), pathToString(v)) for k, v in data)


def symlinkInfoDict(data):
    return _symlinkInfoDict(symlinks(data))


@attribute("realpath")
def realpath(data):
    return functools.partial(_realpathImpl, dict(symlinks(data)))


def _realpathImpl(symlink_table, location):
    location = list(location)
    result = ()
    location_length_when_symlink_encountered = {}

    while location:
        first_element = location.pop(0)
        result = result + (first_element,)
        dest = symlink_table.get(result)
        if dest is not None:
            prevlength = location_length_when_symlink_encountered.get(result)
            if prevlength is not None and len(location) >= prevlength:
                # We encountered some sort of cycle
                return
            location_length_when_symlink_encountered[result] = len(location)
            location = list(dest) + location
            result = ()

    return result


@attribute("_realpaths")
def _realpaths(data):
    rp = functools.partial(_realpathImpl, dict(data))
    return ImmutableList(
        (location, real_destination)
        for location, real_destination in (
            (location, rp(destination)) for location, destination in data
        )
        if real_destination is not None
    )


def realpaths(data):
    return _realpaths(symlinks(data))


@attribute("_resolveStep")
def _resolveStep(data):
    # remove all symlinks that resolve back to a ancestor of its
    # own location
    real_paths = [
        (loc, dest)
        for loc, dest in realpaths(data)
        if loc[0 : len(dest)] != dest
    ]

    if not real_paths:
        return data

    return _resolveImpl.uncached(
        (
            data,
            *(
                i
                for loc, dest in real_paths
                for i in (loc, getSubtree(data, dest))
            ),
        )
    )


@attribute("_resolveStepBack")
def _resolveStepBack(data):
    return _resolveImpl.uncached(
        (
            data,
            *(
                i
                for loc, dest in realpaths(data)
                for i in (loc, getSubtree(data, dest))
            ),
        )
    )


def resolved(data, *, max_steps=5, max_backresolve_steps=1):
    for level in range(max_steps):
        next_data = (
            _resolveStepBack(data)
            if level < max_backresolve_steps
            else _resolveStep(data)
        )
        if data is next_data:
            break
        else:
            data = next_data
    return data


@attribute("_resolveImpl")
def _resolveImpl(resolve_data):
    if len(resolve_data) == 3 and resolve_data[1] == ():
        return resolve_data[2]

    data = resolve_data[0]
    replacements = [
        (resolve_data[i], resolve_data[i + 1])
        for i in range(1, len(resolve_data), 2)
    ]

    if not replacements:
        return data

    try:
        keep_alive = data.meta["_resolveImpl-keep_alive"]
    except KeyError:
        keep_alive = data.meta["_resolveImpl-keep_alive"] = []

    keep_alive = []
    i = 0
    while i < len(replacements):
        first_elem = replacements[i][0][0]
        j = i + 1
        while j < len(replacements):
            if replacements[j][0][0] == first_elem:
                j += 1
            else:
                break

        if j - i == 1 and len(replacements[i][0]) == 1:
            data = _set_helper(data, first_elem, replacements[i][1])
        else:
            subreplacements = ImmutableList(
                [
                    _get_helper(data, first_elem),
                    *(
                        item
                        for k in range(i, j)
                        for item in (
                            replacements[k][0][1:],
                            replacements[k][1],
                        )
                    ),
                ]
            )
            keep_alive.append(subreplacements)
            data = _set_helper(data, first_elem, _resolveImpl(subreplacements))
        i = j

    return data


def _get_helper(data, key):
    try:
        return data[key]
    except Exception:
        return Undefined


def _set_helper(data, key, value):
    if value is Undefined:
        if type(data) is ImmutableDict:
            return data.discard(key)
        else:
            if 0 <= key < len(data):
                return data[0:key] + data[key + 1 :]
            else:
                return data
    else:
        return data.set(key, value)



import builtins
import functools
import inspect
import traceback

from pykzee.core.common import call_soon, Undefined, pathToString
from pykzee.core.Plugin import Plugin


environment = {
    key: value
    for key, value in builtins.__dict__.items()
    if key
    not in (
        "__dict__",
        "__import__",
        "__loader__",
        "__name__",
        "__package__",
        "__spec__",
        "eval",
        "exec",
        "exit",
        "open",
    )
}
environment.update(Undefined=Undefined, call_soon=call_soon)


class CodePlugin(Plugin):
    def init(self, config):
        try:
            code = compile(
                config["code.py"],
                f"<{ pathToString(self.path + ('code.py',)) }>",
                "exec",
            )
            globals = {
                "__builtins__": dict(
                    environment,
                    path=self.path,
                    get=self.get,
                    subscribe=self.subscribe,
                    command=self.command,
                    set_state=self.set,
                    register_command=self.registerCommand,
                    state_from_subscription=self.stateFromSubscription,
                )
            }
            exec(code, globals)
        except Exception as ex:
            self.set(
                (), {"exception": str(ex), "traceback": traceback.format_exc()}
            )

    def stateFromSubscription(self, handler, *paths):
        handler = functools.partial(self.subscriptionCallback, handler)
        if paths:
            self.set((), "Waiting for subscription...")
            return self.subscribe(handler, *paths)
        else:
            call_soon(handler)
            return lambda: None

    async def subscriptionCallback(self, handler, *state):
        try:
            ret = handler(*state)
            if inspect.isawaitable(ret):
                ret = await ret
            self.set((), ret)
        except Exception as ex:
            self.set(
                (), {"exception": str(ex), "traceback": traceback.format_exc()}
            )



import asyncio
from collections.abc import Mapping, Sequence
import enum
import inspect
import re
import typing
import urllib.parse

from pyimmutable import ImmutableDict, ImmutableList


__all__ = (
    "Undefined PathType InvalidPathElement PathElementTypeMismatch "
    "sanitize getDataForPath setDataForPath "
    "makePath stringToPathElement pathToString "
    "waitForOne call_soon print_exception_task_callback".split()
)


def _make_atom(name, *, bool=None):
    typedict = {"__repr__": lambda self: name}
    if bool is not None:
        typedict["__bool__"] = (
            (lambda self: True) if bool else (lambda self: False)
        )
    atomtype = type(f"{ name }Type", (object,), typedict)
    atom = atomtype()
    atomtype.__new__ = (lambda x: lambda cls: x)(atom)
    return atom


Undefined = _make_atom("Undefined", bool=False)

PathElementType = typing.Union[str, int]
PathType = typing.Tuple[PathElementType]


class StateType(enum.Enum):
    RAW = 0
    UNRESOLVED = 1
    RESOLVED = 2


class InvalidPathElement(Exception):
    def __init__(self, value):
        super(InvalidPathElement, self).__init__(
            f"Path elements must be str or int, not {type(value) !r}"
        )


class PathElementTypeMismatch(Exception):
    def __init__(self, elem, data):
        super(PathElementTypeMismatch, self).__init__(
            f"Path element of type { type(elem) !r} cannot be resolved "
            f"in data of type { type(data) !r}"
        )


def sanitize(data):
    t = type(data)
    if (t is ImmutableList or t is ImmutableDict) and data.isImmutableJson:
        return data
    if data is None or t is bool or t is str or t is int or t is float:
        return data
    elif isinstance(data, Sequence):
        data = ImmutableList(sanitize(x) for x in data if x is not Undefined)
        if data.isImmutableJson:
            return data
    elif isinstance(data, Mapping):
        data = ImmutableDict(
            (enforceKeyType(key), sanitize(value))
            for key, value in data.items()
            if value is not Undefined
        )
        if data.isImmutableJson:
            return data
    raise TypeError("data cannot be sanitized to immutable json")


def enforceKeyType(s):
    if type(s) is not str:
        raise TypeError("Dictionary keys must be strings")
    return s


def getDataForPath(data, path: PathType):
    for p in path:
        if type(data) not in (ImmutableDict, ImmutableList):
            return Undefined
        if type(p) not in (str, int):
            raise InvalidPathElement(p)
        try:
            data = data[p]
        except Exception:
            return Undefined
    return data


def setDataForPath(data, path: PathType, value, *, undefined=ImmutableDict()):
    if not path:
        if value is Undefined:
            return undefined
        else:
            return sanitize(value)
    p, path = path[0], path[1:]
    if type(p) is str:
        if data is Undefined:
            data = ImmutableDict()
        elif type(data) is not ImmutableDict:
            raise PathElementTypeMismatch(p, data)
        if value is Undefined and not path:
            return data.discard(p)
    elif type(p) is int:
        if data is Undefined:
            data = ImmutableList()
        elif type(data) is not ImmutableList:
            raise PathElementTypeMismatch(p, data)
        if value is Undefined and not path:
            if 0 <= p < len(data):
                return data[0:p] + data[p + 1 :]
            else:
                return data
        if p < 0 or p > len(data):
            raise IndexError
        if p == len(data):
            return data.append(setDataForPath(Undefined, path, value))
    else:
        raise InvalidPathElement(p)
    return data.set(p, setDataForPath(data.get(p, Undefined), path, value))


def makePath(
    s: typing.Union[str, typing.Sequence[PathElementType]],
    *,
    relativeTo: PathType = (),
) -> PathType:
    if type(s) is str:
        absolute = s.startswith("/")
        s = s.strip("/")
        if not s:
            return () if absolute else relativeTo
        result = [] if absolute else list(relativeTo)
        for e in s.split("/"):
            if e == "..":
                if result:
                    result.pop()
            elif e != ".":
                result.append(stringToPathElement(e))
        return tuple(result)
    elif isinstance(s, Sequence):
        result = tuple(s)
        for i in result:
            if type(i) not in (str, int):
                raise TypeError(
                    "Path sequence must only contain str and int elements"
                )
        return result
    else:
        raise TypeError(f"Cannot convert type { type(s).__name__ } to path")


_rex_integer_element = re.compile(r"^\[(\d+)\]$")


def stringToPathElement(e: str) -> PathElementType:
    res = _rex_integer_element.match(e)
    if res:
        return int(res.group(1))
    return urllib.parse.unquote(e)


def pathToString(path: PathType) -> str:
    return "/" + "/".join(pathElementToString(e) for e in path)


def pathElementToString(e: PathElementType) -> str:
    if type(e) is int:
        return f"[{ e }]"
    if e == ".":
        return "%2E"
    if e == "..":
        return "%2E."
    e = e.replace("%", "%25").replace("/", "%2F")
    if e.startswith("["):
        return f"%5B{ e[1:] }"
    return e


async def waitForOne(*aws):
    _, pending = await asyncio.wait(aws, return_when=asyncio.FIRST_COMPLETED)
    for f in pending:
        f.cancel()


def call_soon(func, *args, **kwargs):
    async def makeCall():
        ret = func(*args, **kwargs)
        if inspect.isawaitable(ret):
            await ret

    asyncio.get_event_loop().call_soon(
        lambda: asyncio.create_task(makeCall()).add_done_callback(
            print_exception_task_callback
        )
    )


def print_exception_task_callback(task):
    if not task.cancelled():
        ex = task.exception()
        if ex is not None and type(ex) is not KeyboardInterrupt:
            task.print_stack()



import asyncio
import collections
import functools

import importlib
import inspect
import logging
import sys
import traceback

from pyimmutable import ImmutableDict, ImmutableList

from pykzee.core.common import (
    call_soon,
    getDataForPath,
    makePath,
    pathToString,
    PathType,
    print_exception_task_callback,
    sanitize,
    setDataForPath,
    StateType,
    Undefined,
)
from pykzee.core import AttachedInfo


SubscriptionSlot = collections.namedtuple(
    "SubscriptionSlot", ("path", "directory", "stateType")
)


class Subscription:
    __slots__ = (
        "plugin",
        "slots",
        "callback",
        "__currentState",
        "__reportedState",
        "disabled",
    )

    def __init__(self, plugin, slots, callback, state, initial: bool):
        if (
            type(slots) != tuple
            or any(type(slot) is not SubscriptionSlot for slot in slots)
            or type(state) != ImmutableList
            or not len(slots) == len(state)
        ):
            logging.error(  # FIXME REMOVE
                "Subscription constructor called with invalid arguments: "
                f"slots={ slots !r} len(state)={ len(state) }"
            )
            raise Exception(
                "Subscription constructor called with invalid arguments: "
                f"slots={ slots !r} len(state)={ len(state) }"
            )
        self.plugin = plugin
        self.slots = slots
        self.callback = callback
        self.__currentState = state
        self.__reportedState = (
            ImmutableList(Undefined for _ in slots) if initial else state
        )
        self.disabled = False

    def setCurrentState(self, idx, state):
        old_state = self.__currentState
        self.__currentState = self.__currentState.set(idx, state)
        return self.__currentState is not old_state

    def getState(self):
        return self.__currentState

    def update(self):
        if (
            not self.disabled
            and self.__reportedState is not self.__currentState
        ):
            self.__reportedState = self.__currentState
            call_soon(self.callback, *self.__currentState)


class Directory:
    __slots__ = (
        "parent",
        "pathElement",
        "subdirectories",
        "subscriptions",
        "state",
    )

    def __init__(self, parent, path_element):
        self.parent = parent
        self.pathElement = path_element
        self.subdirectories = {}
        self.subscriptions = set()  # (sub, idx) tuples
        self.state = Undefined
        try:
            if type(parent.state) in (ImmutableDict, ImmutableList):
                self.state = parent.state[path_element]
        except Exception:
            ...

    def get(self, path: PathType, *, create=True):
        d = self
        for p in path:
            sd = d.subdirectories.get(p)
            if sd is None:
                if create:
                    sd = d.subdirectories[p] = type(self)(d, p)
                else:
                    return
            d = sd
        return d

    def garbageCollect(self):
        parent = self.parent
        if (
            parent is not None
            and not self.subdirectories
            and not self.subscriptions
        ):
            del parent.subdirectories[self.pathElement]
            self.parent = None
            parent.garbageCollect()

    def update(self, new_state, updated_subscriptions):
        if new_state is self.state:
            return

        for sub, idx in self.subscriptions:
            if sub.setCurrentState(idx, new_state):
                updated_subscriptions.add(sub)

        for key, subdir in self.subdirectories.items():
            sdata = Undefined
            if type(new_state) in (ImmutableDict, ImmutableList):
                try:
                    sdata = new_state[key]
                except Exception:
                    ...

            subdir.update(sdata, updated_subscriptions)

        self.state = new_state


class Command:
    __slots__ = "path", "name", "function", "doc", "plugin", "disabled"

    def __init__(self, path, name, function, doc, plugin):
        self.path = path
        self.name = name
        self.function = function
        self.doc = doc
        self.plugin = plugin
        self.disabled = False


class PluginInfo:
    __slots__ = (
        "path",
        "configuration",
        "plugin_object",
        "state",
        "subscriptions",
        "registeredCommands",
        "disabled",
    )

    def __init__(self, *, path, configuration):
        self.path = path
        self.configuration = configuration
        self.plugin_object = None
        self.state = None
        self.subscriptions = set()
        self.registeredCommands = set()
        self.disabled = False


class ManagedTree:
    __slots__ = """
    __rawState __unresolvedState __resolvedState __nextState __realpath
    __rawSubscriptionRoot __unresolvedSubscriptionRoot
    __resolvedSubscriptionRoot __updatedSubscriptions
    __pluginInfos __pluginList __coreState __corePluginPaths
    __commands
    __stateUpdateEvent __stateUpdateTask
    """.strip().split()

    def __init__(self):
        self.__rawState = self.__unresolvedState = ImmutableDict()
        self.__resolvedState = self.__nextState = ImmutableDict()
        self.__realpath = makePath
        self.__rawSubscriptionRoot = Directory(None, None)
        self.__unresolvedSubscriptionRoot = Directory(None, None)
        self.__resolvedSubscriptionRoot = Directory(None, None)
        self.__updatedSubscriptions = set()
        self.__pluginInfos = []
        self.__pluginList = ImmutableList()
        self.__coreState = ImmutableDict(commands=ImmutableDict())
        self.__corePluginPaths = []
        self.__commands = {}  # path -> {name: Command}
        self.__stateUpdateEvent = asyncio.Event()
        self.__stateUpdateTask = asyncio.create_task(
            self.__stateUpdateTaskImpl()
        )
        self.__stateUpdateTask.add_done_callback(print_exception_task_callback)

    def get(self, path: PathType, *, type: StateType = StateType.RESOLVED):
        if type == StateType.RAW:
            state = self.__rawState
        elif type == StateType.UNRESOLVED:
            state = self.__unresolvedState
        else:
            state = self.__resolvedState

        return getDataForPath(state, makePath(path))

    def setRawState(self, new_state: ImmutableDict):
        new_state = sanitize(new_state)
        if self.__rawState is new_state:
            return
        if not new_state.isImmutableJson:
            raise Exception("Invalid state (not immutable json)")
        self.__rawState = new_state
        self.__updatePlugins()

        for plugin in self.__pluginInfos:
            new_state = setDataForPath(new_state, plugin.path, plugin.state)

        self.__nextState = new_state
        self.__setCore(
            ("plugins",),
            {pathToString(path): config for path, config in self.__pluginList},
            force=True,
        )

    def __updatePlugins(self):
        new_plugin_list = AttachedInfo.plugins(self.__rawState)
        if new_plugin_list is self.__pluginList:
            return

        old_plugin_infos = self.__pluginInfos
        new_plugin_infos = []
        core_plugin_paths = []
        old_index = new_index = 0

        while True:
            have_old = old_index < len(old_plugin_infos)
            have_new = new_index < len(new_plugin_list)

            if not (have_new or have_old):
                break

            if have_new:
                npath, nconfig = new_plugin_list[new_index]
                if nconfig["__plugin__"] == "core-plugin":
                    core_plugin_paths.append(npath)
                    new_index += 1
                    continue

            if have_old:
                opi = old_plugin_infos[old_index]

            if not have_new or (have_old and opi.path < npath):
                self.__removePlugin(opi)
                old_index += 1
            elif not have_old or (have_new and npath < opi.path):
                new_plugin_infos.append(self.__newPlugin(npath, nconfig))
                new_index += 1
            else:
                old_index += 1
                new_index += 1
                new_plugin_infos.append(self.__updatePlugin(opi, nconfig))

        self.__pluginInfos = new_plugin_infos
        self.__pluginList = new_plugin_list
        self.__corePluginPaths = core_plugin_paths

    def __removePlugin(self, plugin_info):
        plugin_info.disabled = True
        plugin_object = plugin_info.plugin_object
        subscriptions = plugin_info.subscriptions
        registered_commands = plugin_info.registeredCommands

        plugin_info.configuration = None
        plugin_info.plugin_object = None
        plugin_info.state = None
        plugin_info.subscriptions = set()
        plugin_info.registeredCommands = set()

        for sub in subscriptions:
            self.unsubscribe(sub)

        for cmd in registered_commands:
            self.unregisterCommand(cmd)

        try:
            plugin_object.shutdown()
        except Exception:
            ...

    def __newPlugin(self, path, config):
        plugin_info = PluginInfo(path=path, configuration=config)

        try:
            plugin_identifier = config["__plugin__"]
            module, class_ = plugin_identifier.rsplit(".", 1)

            sys.modules.pop(module, None)
            mod = importlib.import_module(module)

            PluginType = getattr(mod, class_)
            plugin_info.plugin_object = PluginType(
                path=path,
                get=lambda path: self.get(path),
                subscribe=lambda callback, *paths, initial=True: (
                    self.subscribe(
                        plugin_info, paths, callback, initial=initial
                    )
                ),
                command=self.command,
                set_state=functools.partial(
                    self.__setPluginState, plugin_info
                ),
                register_command=functools.partial(
                    self.registerCommand, plugin_info
                ),
            )
            plugin_info.plugin_object.init(config)
        except Exception as ex:
            traceback.print_exc()
            plugin_info.state = ImmutableDict(
                exception=str(ex), traceback=traceback.format_exc()
            )
            plugin_info.plugin_object = None

        return plugin_info

    def __updatePlugin(self, plugin_info, new_config):
        if plugin_info.configuration is new_config:
            return plugin_info

        try:
            if (
                plugin_info.configuration["__plugin__"]
                == new_config["__plugin__"]
                and plugin_info.plugin_object
                and hasattr(plugin_info.plugin_object, "updateConfig")
                and plugin_info.plugin_object.updateConfig(new_config)
            ):
                plugin_info.configuration = new_config
                return plugin_info
        except Exception:
            traceback.print_exc()

        self.__removePlugin(plugin_info)
        return self.__newPlugin(plugin_info.path, new_config)

    def __set(self, path: PathType, value):
        self.__nextState = setDataForPath(self.__nextState, path, value)
        if self.__nextState is not self.__unresolvedState:
            self.__stateUpdateEvent.set()

    def __setCore(self, path, value, *, force=True):
        new_core_state = setDataForPath(self.__coreState, path, value)
        if self.__coreState is new_core_state and not force:
            return
        self.__coreState = new_core_state
        next_state = self.__nextState
        for core_plugin_path in self.__corePluginPaths:
            next_state = setDataForPath(
                next_state, core_plugin_path, new_core_state
            )
        if next_state is not self.__nextState:
            self.__nextState = next_state
            self.__stateUpdateEvent.set()

    def __setPluginState(self, plugin_info, path, value):
        if not plugin_info.disabled:
            new_state = setDataForPath(
                plugin_info.state, path, value, undefined=None
            )
            if plugin_info.state is not new_state:
                plugin_info.state = new_state
                self.__set(plugin_info.path, new_state)

    def command(self, path, cmd):
        return self.__commands[makePath(path)][cmd].function

    def subscribe(
        self,
        plugin_info,
        paths,
        callback,
        *,
        initial=True,
        state_type=StateType.RESOLVED,
    ):
        if plugin_info.disabled:
            raise Exception("disabled plugin must not subscribe")
        slots = tuple(
            SubscriptionSlot(
                path,
                (
                    self.__rawSubscriptionRoot
                    if state_type == StateType.RAW
                    else self.__unresolvedSubscriptionRoot
                    if state_type == StateType.UNRESOLVED
                    else self.__resolvedSubscriptionRoot
                ).get(path),
                state_type,
            )
            for path in map(makePath, paths)
        )
        state = ImmutableList(slot.directory.state for slot in slots)
        sub = Subscription(plugin_info, slots, callback, state, initial)
        plugin_info.subscriptions.add(sub)
        for idx, slot in enumerate(slots):
            slot.directory.subscriptions.add((sub, idx))
        if initial:
            self.__updatedSubscriptions.add(sub)
            self.__stateUpdateEvent.set()
        return lambda: self.unsubscribe(sub)

    def unsubscribe(self, sub):
        sub.disabled = True
        for idx, slot in enumerate(sub.slots):
            slot.directory.subscriptions.discard((sub, idx))
            slot.directory.garbageCollect()
        sub.plugin.subscriptions.discard(sub)

    def registerCommand(
        self, plugin_info, path, name, function, *, doc=Undefined
    ):
        if plugin_info.disabled:
            raise Exception("Disabled plugins cannot register commands")
        path = plugin_info.path + makePath(path)
        if doc is Undefined:
            doc = function.__doc__
        sig = inspect.signature(function)
        try:
            path_commands = self.__commands[path]
        except KeyError:
            path_commands = self.__commands[path] = {}
        if name in path_commands:
            raise Exception("Command { path }:{ name } already registered")
        cmd = Command(path, name, function, doc, plugin_info)
        plugin_info.registeredCommands.add(cmd)
        path_commands[name] = cmd

        self.__setCore(
            ("commands", pathToString(path), name),
            {"doc": doc, "signature": str(sig)},
        )

        return lambda: self.__unregisterCommand(cmd)

    def unregisterCommand(self, cmd):
        if cmd.disabled:
            return
        cmd.disabled = True
        path_commands = self.__commands[cmd.path]
        path_commands.pop(cmd.name)
        cmd.plugin.registeredCommands.discard(cmd)
        if not path_commands:
            del self.__commands[cmd.path]
            self.__setCore(("commands", pathToString(cmd.path)), Undefined)
        else:
            self.__setCore(
                ("commands", pathToString(cmd.path), cmd.name), Undefined
            )

    async def __stateUpdateTaskImpl(self):
        while True:
            await self.__stateUpdateEvent.wait()
            self.__stateUpdateEvent.clear()

            next_state = self.__nextState
            if self.__unresolvedState is not next_state:
                for path_prefix in self.__corePluginPaths:
                    next_state = setDataForPath(
                        next_state,
                        path_prefix + ("symlinks",),
                        AttachedInfo.symlinkInfoDict(next_state),
                    )

                self.__resolvedState = AttachedInfo.resolved(next_state)
                self.__realpath = (
                    lambda func: lambda path: func(makePath(path))
                )(AttachedInfo.realpath(next_state))
                self.__unresolvedState = self.__nextState = next_state

            self.__rawSubscriptionRoot.update(
                self.__rawState, self.__updatedSubscriptions
            )
            self.__unresolvedSubscriptionRoot.update(
                self.__unresolvedState, self.__updatedSubscriptions
            )
            self.__resolvedSubscriptionRoot.update(
                self.__resolvedState, self.__updatedSubscriptions
            )

            updated_subscriptions = self.__updatedSubscriptions
            self.__updatedSubscriptions = set()
            for sub in updated_subscriptions:
                sub.update()



import asyncio
import json
import logging
import os
import stat
import traceback

import aiofiles
from pyimmutable import ImmutableDict
import watchdog.events
import watchdog.observers

from pykzee.core.common import print_exception_task_callback


class RawStateLoader:
    def __init__(self, set_raw_state):
        self.__setRawState = set_raw_state
        self.__shutdown = asyncio.Future()
        self.__reread_tree_event = asyncio.Event()

        asyncio.create_task(self.__rereadTaskImpl()).add_done_callback(
            print_exception_task_callback
        )

        observer = self.__observer = watchdog.observers.Observer()
        loop = asyncio.get_event_loop()
        observer.schedule(
            WatchdogEventHandler(
                lambda: loop.call_soon_threadsafe(self.__reread_tree_event.set)
            ),
            ".",
            recursive=True,
        )
        observer.start()

    def __del__(self):
        self.__observer.stop()
        self.__observer.join()

    async def readStateFromDisk(self):
        self.__reread_tree_event.clear()
        self.__setRawState(
            ImmutableDict([x async for x in load_state_tree(".")])
        )

    async def run(self):
        return await self.__shutdown

    async def __rereadTaskImpl(self):
        while True:
            await self.__reread_tree_event.wait()
            await asyncio.sleep(2)
            if self.__reread_tree_event.is_set():
                try:
                    await self.readStateFromDisk()
                except Exception:
                    traceback.print_exc()


class WatchdogEventHandler(watchdog.events.FileSystemEventHandler):
    def __init__(self, callback):
        self.__callback = callback

    def on_any_event(self, event):
        self.__callback()


async def load_state_tree(dirpath):
    for filename in sorted(os.listdir(dirpath)):
        if filename.startswith(".") or filename.endswith("~"):
            continue

        key = filename
        fspath = os.path.join(dirpath, filename)
        mode = os.stat(fspath).st_mode

        if stat.S_ISDIR(mode):
            yield key, ImmutableDict(
                [x async for x in load_state_tree(fspath)]
            )
        elif stat.S_ISREG(mode):
            async with aiofiles.open(fspath) as f:
                content = await f.read()
            if key.endswith(".json"):
                key = key[:-5]
                content = json.loads(content)
            elif key.endswith(".txt"):
                key = key[:-4]
            yield key, content
        else:
            logging.warning(
                f"ConfigPlugin: ignoring non-regular file f{ fspath }"
            )



from setuptools import setup

with open("README.rst", "r") as fh:
    long_description = fh.read()

with open("pykzee/core/__version__.py", "r") as fh:
    versiondict = {"__builtins__": {}}
    exec(fh.read(), versiondict)
    version = versiondict["version"]

setup(
    name="pykzee",
    version=version,
    description=(
        "Core engine allowing plug-ins to operate on a JSON-like state tree"
    ),
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/spacedentist/pykzee",
    download_url=(
        f"https://github.com/spacedentist/pykzee/archive/{ version }.tar.gz"
    ),
    author="Sven Over",
    author_email="sp@cedenti.st",
    license="MIT",
    packages=["pykzee", "pykzee.core"],
    install_requires=[
        "aiofiles>=0.4.0",
        "watchdog>=0.9.0",
        "pyimmutable>=0.2.0",
    ],
    entry_points={"console_scripts": ["pykzee=pykzee.core.__main__:main"]},
    test_suite="pykzee.core.tests",
)



import collections

from pyimmutable import ImmutableDict
from pykzee.core.common import Undefined, setDataForPath


class Tree:
    class RegisteredCommand:
        __slots__ = "function", "doc", "unregister", "disabled"

        def __init__(self, function, doc, unregister):
            self.function = function
            self.doc = doc
            self.unregister = unregister
            self.disabled = False

    TreeAccess = collections.namedtuple(
        "TreeAccess",
        (
            "set",
            "submitState",
            "registerCommand",
            "createSubtree",
            "clear",
            "deactivate",
        ),
    )

    def __init__(
        self,
        path,
        *,
        parent_set,
        parent_register_command,
        immediate_updates=True
    ):
        self.__path = path
        self.__parentSet = parent_set
        self.__parentRegisterCommand = parent_register_command
        self.__state = ImmutableDict()
        self.__reportedState = Undefined
        self.__registeredCommands = {}
        self.__immediate_updates = immediate_updates
        self.__deactivated = False
        self.__hidden = False

    @property
    def path(self):
        return self.__path

    def getAccessProxy(self):
        return self.TreeAccess(
            self.set,
            self.submitState,
            self.registerCommand,
            self.createSubtree,
            self.clear,
            self.deactivate,
        )

    def set(self, path, value):
        self.__state = setDataForPath(self.__state, path, value)
        if self.__immediate_updates:
            self.submitState()

    def registerCommand(self, path, name, function, *, doc=Undefined):
        if doc is Undefined:
            doc = function.__doc__
        existing_rc = self.__registeredCommands.get((path, name))
        if existing_rc is not None:
            existing_rc.disabled = True
            existing_rc.unregister()
        if self.__hidden:
            unregister = no_op
        else:
            unregister = self.__parentRegisterCommand(
                self.__path + path, name, function, doc=doc
            )
        rc = self.RegisteredCommand(function, doc, unregister)
        self.__registeredCommands[path, name] = rc

        def unregister_command():
            if not rc.disabled:
                rc.disabled = True
                del self.__registeredCommands[path, name]
                rc.unregister()

        return unregister_command

    def createSubtree(self, path, *, immediate_updates=True):
        return Tree(
            path,
            parent_set=self.set,
            parent_register_command=self.registerCommand,
            immediate_updates=immediate_updates,
        )

    def clear(self):
        for rc in self.__registeredCommands.values():
            rc.disabled = True
            rc.unregister()
        self.__registeredCommands = {}
        self.__state = ImmutableDict()
        self.__reportedState = Undefined
        self.__parentSet(self.__path, Undefined)

    def deactivate(self):
        if not self.__deactivated:
            self.clear()
            self.__parentSet = self.__parentRegisterCommand = raise_deactivated
            self.__deactivated = True

    def hide(self):
        if self.__hidden:
            return
        for rc in self.__registeredCommands.values():
            rc.unregister()
            rc.unregister = no_op
        self.__parentSet(self.__path, Undefined)
        self.__hidden = True

    def show(self, new_path=None):
        if not self.__hidden and (new_path is None or new_path == self.__path):
            return
        self.hide()
        if new_path is not None:
            self.__path = new_path
        self.__parentSet(self.__path, self.__state)
        self.__hidden = False
        self.__reportedState = self.__state
        for (path, name), rc in self.__registeredCommands.items():
            rc.unregister = self.__parentRegisterCommand(
                self.__path + path, name, rc.function, rc.doc
            )

    def submitState(self):
        if self.__reportedState is not self.__state and not self.__hidden:
            self.__parentSet(self.__path, self.__state)
            self.__reportedState = self.__state


def raise_deactivated(*args, **kwargs):
    raise Exception("Subtree has been deactivated")


def no_op():
    ...



import argparse
import asyncio
import logging
import os


from pykzee.core.RawStateLoader import RawStateLoader
from pykzee.core.ManagedTree import ManagedTree

logging.getLogger().setLevel(logging.DEBUG)

try:
    import coloredlogs
except ImportError:
    ...
else:
    coloredlogs.install(level="DEBUG")


parser = argparse.ArgumentParser()
parser.add_argument(
    "--config",
    help="path to config directory (defaults to current working directory)",
)
options = parser.parse_args()


async def amain():
    if options.config:
        os.chdir(options.config)

    mtree = ManagedTree()
    raw_state_loader = RawStateLoader(mtree.setRawState)
    await raw_state_loader.readStateFromDisk()
    await raw_state_loader.run()


def main():
    asyncio.run(amain())


if __name__ == "__main__":
    main()



import logging

from pyimmutable import ImmutableDict, ImmutableList
from pykzee.core.common import makePath
from pykzee.core.Plugin import Plugin


class StateLoggerPlugin(Plugin):
    def init(self, config):
        self.__pretty = bool(config.get("pretty", False))
        self.unsubscribe = self.subscribe(
            self.stateUpdate, makePath(config.get("path", ()))
        )

    def updateConfig(self, new_config):
        self.unsubscribe()
        self.__pretty = new_config["pretty"]
        self.unsubscribe = self.subscribe(
            self.stateUpdate, new_config.get("path", ())
        )
        return True

    def stateUpdate(self, state):
        if not self.__pretty:
            logging.debug(repr(state))
            return

        logging.debug("StateLoggerPlugin: new state:")
        pretty_print(state, OutputLines(logging.debug))


class OutputLines:
    def __init__(self, write):
        self.__write = write
        self.__data = ""

    def __call__(self, x):
        self.__data += x
        pos = self.__data.rfind("\n")
        if pos < 0:
            return
        out = self.__data[0:pos]
        self.__data = self.__data[pos + 1 :]
        for line in out.split("\n"):
            self.__write(line)

    def __del__(self):
        if self.__data:
            self.__write(self.__data)


def pretty_print(data, write, indent=""):
    if type(data) is ImmutableDict:
        more_indent = indent + "  "
        write("{\n")
        for key, value in data.items():
            write(f"{ indent }  { key !r}: ")
            pretty_print(value, write, more_indent)
            write(",\n")
        write(f"{ indent }}}")
    elif type(data) is ImmutableList:
        more_indent = indent + "  "
        write("[\n")
        for value in data:
            write(more_indent)
            pretty_print(value, write, more_indent)
            write(",\n")
        write(f"{ indent }]")
    else:
        write(repr(data))
